@@ -1,27 +1,6 @@
-/*=========================================================================
-
-  Program:   Visualization Toolkit
-  Module:    vtkMath.h
-
-  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
-  All rights reserved.
-  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.
-
-     This software is distributed WITHOUT ANY WARRANTY; without even
-     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
-     PURPOSE.  See the above copyright notice for more information.
-
-=========================================================================
-  Copyright 2011 Sandia Corporation.
-  Under the terms of Contract DE-AC04-94AL85000, there is a non-exclusive
-  license for use of this work by or on behalf of the
-  U.S. Government. Redistribution and use in source and binary forms, with
-  or without modification, are permitted provided that this Notice and any
-  statement of authorship are reproduced on all copies.
-
-  Contact: pppebay@sandia.gov,dcthomp@sandia.gov
-
-=========================================================================*/
+// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
+// SPDX-FileCopyrightText: Copyright 2011 Sandia Corporation
+// SPDX-License-Identifier: LicenseRef-BSD-3-Clause-Sandia-USGov
 /**
  * @class   vtkMath
  * @brief   performs common math operations
@@ -48,8 +27,9 @@
 
 #include "vtkMathConfigure.h" // For <cmath> and VTK_HAS_ISNAN etc.
 
-#include <algorithm> // for std::clamp
-#include <cassert>   // assert() in inline implementations.
+#include <algorithm>   // for std::clamp
+#include <cassert>     // assert() in inline implementations.
+#include <type_traits> // for type_traits
 
 #ifndef DBL_MIN
 #define VTK_DBL_MIN 2.2250738585072014e-308
@@ -71,19 +51,24 @@
 #endif // DBL_EPSILON
 #endif // VTK_DBL_EPSILON
 
+VTK_ABI_NAMESPACE_BEGIN
 class vtkDataArray;
 class vtkPoints;
 class vtkMathInternal;
 class vtkMinimalStandardRandomSequence;
 class vtkBoxMuellerRandomSequence;
+VTK_ABI_NAMESPACE_END
 
 namespace vtk_detail
 {
+VTK_ABI_NAMESPACE_BEGIN
 // forward declaration
 template <typename OutT>
 void RoundDoubleToIntegralIfNecessary(double val, OutT* ret);
+VTK_ABI_NAMESPACE_END
 } // end namespace vtk_detail
 
+VTK_ABI_NAMESPACE_BEGIN
 class VTKCOMMONCORE_EXPORT vtkMath : public vtkObject
 {
 public:
@@ -91,6 +76,36 @@ public:
   vtkTypeMacro(vtkMath, vtkObject);
   void PrintSelf(ostream& os, vtkIndent indent) override;
 
+private:
+  template <class VectorT, class = void>
+  struct VectorImplementsSize : std::false_type
+  {
+  };
+
+  template <class VectorT>
+  struct VectorImplementsSize<VectorT, decltype((void)std::declval<VectorT>().size(), void())>
+    : std::true_type
+  {
+  };
+
+  /**
+   * Convenient meta-class to enable function signatures if `VectorT::size()` exists;
+   */
+  template <class VectorT>
+  using EnableIfVectorImplementsSize =
+    typename std::enable_if<VectorImplementsSize<VectorT>::value>::type;
+
+public:
+  /**
+   * When this value is passed to a select templated functions in `vtkMath`,
+   * the computation can be performed on dynamic sized arrays as long as they implement the method
+   * `size()`. One can also pass this constant to `vtkDataArrayTupleRange` to get dynamic sized
+   * tuples.
+   *
+   * @sa vtkDataArrayRange
+   */
+  static constexpr int DYNAMIC_VECTOR_SIZE() { return 0; }
+
   /**
    * A mathematical constant. This version is atan(1.0) * 4.0
    */
@@ -348,6 +363,20 @@ public:
     }
   }
 
+  /**
+   * Addition of two 3-vectors (double version). Result is stored in c according to c = a + b.
+   *
+   * Each parameter needs to implement `operator[]`
+   */
+  template <class VectorT1, class VectorT2, class VectorT3>
+  static void Add(VectorT1&& a, VectorT2&& b, VectorT3& c)
+  {
+    for (int i = 0; i < 3; ++i)
+    {
+      c[i] = a[i] + b[i];
+    }
+  }
+
   /**
    * Subtraction of two 3-vectors (float version). Result is stored in c according to c = a - b.
    */
@@ -499,6 +528,14 @@ public:
     }
   }
 
+  /**
+   * Cross product of two 3-vectors. Result (a x b) is stored in c.
+   *
+   * Input vectors need to implement operator[]
+   */
+  template <class VectorT1, class VectorT2, class VectorT3>
+  static void Cross(VectorT1&& a, VectorT2&& b, VectorT3& c);
+
   /**
    * Cross product of two 3-vectors. Result (a x b) is stored in c.
    * (float version)
@@ -640,27 +677,27 @@ public:
    * Compute the amplitude of a Gaussian function with mean=0 and specified variance.
    * That is, 1./(std::sqrt(2 Pi * variance)) * exp(-distanceFromMean^2/(2.*variance)).
    */
-  static double GaussianAmplitude(const double variance, const double distanceFromMean);
+  static double GaussianAmplitude(double variance, double distanceFromMean);
 
   /**
    * Compute the amplitude of a Gaussian function with specified mean and variance.
    * That is, 1./(std::sqrt(2 Pi * variance)) * exp(-(position - mean)^2/(2.*variance)).
    */
-  static double GaussianAmplitude(const double mean, const double variance, const double position);
+  static double GaussianAmplitude(double mean, double variance, double position);
 
   /**
    * Compute the amplitude of an unnormalized Gaussian function with mean=0 and specified variance.
    * That is, exp(-distanceFromMean^2/(2.*variance)). When distanceFromMean = 0, this function
    * returns 1.
    */
-  static double GaussianWeight(const double variance, const double distanceFromMean);
+  static double GaussianWeight(double variance, double distanceFromMean);
 
   /**
    * Compute the amplitude of an unnormalized Gaussian function with specified mean and variance.
    * That is, exp(-(position - mean)^2/(2.*variance)). When the distance from 'position' to 'mean'
    * is 0, this function returns 1.
    */
-  static double GaussianWeight(const double mean, const double variance, const double position);
+  static double GaussianWeight(double mean, double variance, double position);
 
   /**
    * Dot product of two 2-vectors. (float version).
@@ -811,9 +848,10 @@ public:
     class LayoutT1 = vtkMatrixUtilities::Layout::Identity,
     class LayoutT2 = vtkMatrixUtilities::Layout::Identity, class MatrixT1, class MatrixT2,
     class MatrixT3>
-  static void MultiplyMatrix(const MatrixT1& M1, const MatrixT2& M2, MatrixT3&& M3)
+  static void MultiplyMatrix(MatrixT1&& M1, MatrixT2&& M2, MatrixT3&& M3)
   {
-    vtkMathPrivate::MultiplyMatrix<RowsT, MidDimT, ColsT, LayoutT1, LayoutT2>::Compute(M1, M2, M3);
+    vtkMathPrivate::MultiplyMatrix<RowsT, MidDimT, ColsT, LayoutT1, LayoutT2>::Compute(
+      std::forward<MatrixT1>(M1), std::forward<MatrixT2>(M2), std::forward<MatrixT3>(M3));
   }
 
   /**
@@ -838,9 +876,10 @@ public:
    */
   template <int RowsT, int ColsT, class LayoutT = vtkMatrixUtilities::Layout::Identity,
     class MatrixT, class VectorT1, class VectorT2>
-  static void MultiplyMatrixWithVector(const MatrixT& M, const VectorT1& X, VectorT2&& Y)
+  static void MultiplyMatrixWithVector(MatrixT&& M, VectorT1&& X, VectorT2&& Y)
   {
-    vtkMathPrivate::MultiplyMatrix<RowsT, ColsT, 1, LayoutT>::Compute(M, X, Y);
+    vtkMathPrivate::MultiplyMatrix<RowsT, ColsT, 1, LayoutT>::Compute(
+      std::forward<MatrixT>(M), std::forward<VectorT1>(X), std::forward<VectorT2>(Y));
   }
 
   /**
@@ -848,11 +887,49 @@ public:
    * VectorT1 and VectorT2 are arrays of size SizeT, and must implement
    * `operator[]`.
    */
-  template <class ScalarT, int SizeT, class VectorT1, class VectorT2>
-  static ScalarT Dot(const VectorT1& x, const VectorT2& y)
+  template <class ScalarT, int SizeT, class VectorT1, class VectorT2,
+    class = typename std::enable_if<SizeT != DYNAMIC_VECTOR_SIZE()>::type>
+  static ScalarT Dot(VectorT1&& x, VectorT2&& y)
   {
     return vtkMathPrivate::ContractRowWithCol<ScalarT, 1, SizeT, 1, 0, 0,
-      vtkMatrixUtilities::Layout::Identity, vtkMatrixUtilities::Layout::Transpose>::Compute(x, y);
+      vtkMatrixUtilities::Layout::Identity,
+      vtkMatrixUtilities::Layout::Transpose>::Compute(std::forward<VectorT1>(x),
+      std::forward<VectorT2>(y));
+  }
+
+  /**
+   * Computes the dot product between 2 vectors x and y.
+   * This function is solely invoked when `SizeT == DYNAMIC_VECTOR_SIZE()`.
+   * VectorT1 and VectorT2 are arrays of dynamic size, and must implement
+   * `operator[]` and `size()`.
+   */
+  template <class ScalarT, int SizeT, class VectorT1, class VectorT2,
+    class = typename std::enable_if<SizeT == DYNAMIC_VECTOR_SIZE()>::type,
+    class = EnableIfVectorImplementsSize<VectorT1>>
+  static ScalarT Dot(VectorT1&& x, VectorT2&& y)
+  {
+    ScalarT dot = 0.0;
+    using SizeType = decltype(std::declval<VectorT1>().size());
+    for (SizeType dim = 0; dim < x.size(); ++dim)
+    {
+      dot += x[dim] * y[dim];
+    }
+    return dot;
+  }
+
+  /**
+   * Computes the dot product between 2 vectors x and y.
+   * VectorT1 and VectorT2 are arrays of size SizeT, and must implement
+   * `operator[]`.
+   *
+   * If `SizeT == DYNAMIC_VECTOR_SIZE()`, then x must implement the method `size()`.
+   */
+  template <int SizeT, class VectorT>
+  static typename vtkMatrixUtilities::ScalarTypeExtractor<VectorT>::value_type SquaredNorm(
+    VectorT&& x)
+  {
+    using Scalar = typename vtkMatrixUtilities::ScalarTypeExtractor<VectorT>::value_type;
+    return vtkMath::Dot<Scalar, SizeT>(std::forward<VectorT>(x), std::forward<VectorT>(x));
   }
 
   /**
@@ -873,9 +950,9 @@ public:
    */
   template <int SizeT, class LayoutT = vtkMatrixUtilities::Layout::Identity, class MatrixT>
   static typename vtkMatrixUtilities::ScalarTypeExtractor<MatrixT>::value_type Determinant(
-    const MatrixT& M)
+    MatrixT&& M)
   {
-    return vtkMathPrivate::Determinant<SizeT, LayoutT>::Compute(M);
+    return vtkMathPrivate::Determinant<SizeT, LayoutT>::Compute(std::forward<MatrixT>(M));
   }
 
   /**
@@ -895,9 +972,10 @@ public:
    */
   template <int SizeT, class LayoutT = vtkMatrixUtilities::Layout::Identity, class MatrixT1,
     class MatrixT2>
-  static void InvertMatrix(const MatrixT1& M1, MatrixT2&& M2)
+  static void InvertMatrix(MatrixT1&& M1, MatrixT2&& M2)
   {
-    vtkMathPrivate::InvertMatrix<SizeT, LayoutT>::Compute(M1, M2);
+    vtkMathPrivate::InvertMatrix<SizeT, LayoutT>::Compute(
+      std::forward<MatrixT1>(M1), std::forward<MatrixT2>(M2));
   }
 
   /**
@@ -915,9 +993,10 @@ public:
    */
   template <int RowsT, int ColsT, class LayoutT = vtkMatrixUtilities::Layout::Identity,
     class MatrixT, class VectorT1, class VectorT2>
-  static void LinearSolve(const MatrixT& M, const VectorT1& x, VectorT2& y)
+  static void LinearSolve(MatrixT&& M, VectorT1&& x, VectorT2&& y)
   {
-    vtkMathPrivate::LinearSolve<RowsT, ColsT, LayoutT>::Compute(M, x, y);
+    vtkMathPrivate::LinearSolve<RowsT, ColsT, LayoutT>::Compute(
+      std::forward<MatrixT>(M), std::forward<VectorT1>(x), std::forward<VectorT2>(y));
   }
 
   /**
@@ -935,13 +1014,16 @@ public:
    * of non zero elements of a diagonal matrix.
    */
   template <class ScalarT, int SizeT, class LayoutT = vtkMatrixUtilities::Layout::Identity,
-    class VectorT1, class MatrixT, class VectorT2>
-  static ScalarT Dot(const VectorT1& x, const MatrixT& M, const VectorT2& y)
+    class VectorT1, class MatrixT, class VectorT2,
+    class = typename std::enable_if<SizeT != DYNAMIC_VECTOR_SIZE()>::type>
+  static ScalarT Dot(VectorT1&& x, MatrixT&& M, VectorT2&& y)
   {
     ScalarT tmp[SizeT];
-    vtkMathPrivate::MultiplyMatrix<SizeT, SizeT, 1, LayoutT>::Compute(M, y, tmp);
+    vtkMathPrivate::MultiplyMatrix<SizeT, SizeT, 1, LayoutT>::Compute(
+      std::forward<MatrixT>(M), std::forward<VectorT2>(y), tmp);
     return vtkMathPrivate::ContractRowWithCol<ScalarT, 1, SizeT, 1, 0, 0,
-      vtkMatrixUtilities::Layout::Identity, vtkMatrixUtilities::Layout::Transpose>::Compute(x, tmp);
+      vtkMatrixUtilities::Layout::Identity,
+      vtkMatrixUtilities::Layout::Transpose>::Compute(std::forward<VectorT1>(x), tmp);
   }
 
   /**
@@ -1017,7 +1099,7 @@ public:
   static void QuaternionToMatrix3x3(const double quat[4], double A[3][3]);
   template <class QuaternionT, class MatrixT,
     class EnableT = typename std::enable_if<!vtkMatrixUtilities::MatrixIs2DArray<MatrixT>()>::type>
-  static void QuaternionToMatrix3x3(const QuaternionT& q, MatrixT&& A);
+  static void QuaternionToMatrix3x3(QuaternionT&& q, MatrixT&& A);
   ///@}
 
   ///@{
@@ -1034,7 +1116,7 @@ public:
   static void Matrix3x3ToQuaternion(const double A[3][3], double quat[4]);
   template <class MatrixT, class QuaternionT,
     class EnableT = typename std::enable_if<!vtkMatrixUtilities::MatrixIs2DArray<MatrixT>()>::type>
-  static void Matrix3x3ToQuaternion(const MatrixT& A, QuaternionT&& q);
+  static void Matrix3x3ToQuaternion(MatrixT&& A, QuaternionT&& q);
   ///@}
 
   ///@{
@@ -1534,6 +1616,16 @@ public:
    */
   static int QuadraticRoot(double a, double b, double c, double min, double max, double* u);
 
+  /**
+   * Compute the greatest common divisor (GCD) of two positive integers m and n. If the
+   * computed GCD==1, then the two integers are coprime to one another. This is a simple,
+   * recursive implementation.
+   */
+  static vtkIdType ComputeGCD(vtkIdType m, vtkIdType n) { return (n ? ComputeGCD(n, m % n) : m); }
+
+  /**
+   * Support the convolution operations.
+   */
   enum class ConvolutionMode
   {
     FULL,
@@ -1582,11 +1674,11 @@ public:
     switch (mode)
     {
       case ConvolutionMode::SAME:
-        begin = static_cast<int>(std::ceil(std::min(sampleSize, kernelSize) / 2.0)) - 1;
-        end = begin + std::max(sampleSize, kernelSize);
+        begin = static_cast<int>(std::ceil((std::min)(sampleSize, kernelSize) / 2.0)) - 1;
+        end = begin + (std::max)(sampleSize, kernelSize);
         break;
       case ConvolutionMode::VALID:
-        begin = std::min(sampleSize, kernelSize) - 1;
+        begin = (std::min)(sampleSize, kernelSize) - 1;
         end = begin + std::abs(sampleSize - kernelSize) + 1;
         break;
       case ConvolutionMode::FULL:
@@ -1598,13 +1690,26 @@ public:
     {
       Iter3 out = beginOut + i - begin;
       *out = 0;
-      for (int j = std::max(i - sampleSize + 1, 0); j <= std::min(i, kernelSize - 1); j++)
+      for (int j = (std::max)(i - sampleSize + 1, 0); j <= (std::min)(i, kernelSize - 1); j++)
       {
         *out += *(beginSample + (i - j)) * *(beginKernel + j);
       }
     }
   }
 
+  /**
+   * Get the coordinates of a point along a line defined by p1 and p2, at a
+   * specified offset relative to p2.
+   */
+  static void GetPointAlongLine(double result[3], double p1[3], double p2[3], const double offset)
+  {
+    double directionVector[3] = { p2[0] - p1[0], p2[1] - p1[1], p2[2] - p1[2] };
+    vtkMath::Normalize(directionVector);
+    result[0] = p2[0] + (offset * directionVector[0]);
+    result[1] = p2[1] + (offset * directionVector[1]);
+    result[2] = p2[2] + (offset * directionVector[2]);
+  }
+
 protected:
   vtkMath() = default;
   ~vtkMath() override = default;
@@ -1791,6 +1896,19 @@ inline ReturnTypeT vtkMath::Distance2BetweenPoints(const TupleRangeT1& p1, const
     (p1[2] - p2[2]) * (p1[2] - p2[2]));
 }
 
+//----------------------------------------------------------------------------
+template <class VectorT1, class VectorT2, class VectorT3>
+void vtkMath::Cross(VectorT1&& a, VectorT2&& b, VectorT3& c)
+{
+  typedef typename vtkMatrixUtilities::ScalarTypeExtractor<VectorT3>::value_type ValueType;
+  ValueType Cx = a[1] * b[2] - a[2] * b[1];
+  ValueType Cy = a[2] * b[0] - a[0] * b[2];
+  ValueType Cz = a[0] * b[1] - a[1] * b[0];
+  c[0] = Cx;
+  c[1] = Cy;
+  c[2] = Cz;
+}
+
 //----------------------------------------------------------------------------
 // Cross product of two 3-vectors. Result (a x b) is stored in c[3].
 inline void vtkMath::Cross(const float a[3], const float b[3], float c[3])
@@ -1922,11 +2040,12 @@ inline void vtkMath::TensorFromSymmetricTensor(T tensor[9])
   tensor[2] = tensor[6]; // XZ
   tensor[1] = tensor[3]; // XY
 }
+VTK_ABI_NAMESPACE_END
 
 namespace
 {
 template <class QuaternionT, class MatrixT>
-inline void vtkQuaternionToMatrix3x3(const QuaternionT& quat, MatrixT& A)
+inline void vtkQuaternionToMatrix3x3(QuaternionT&& quat, MatrixT&& A)
 {
   typedef typename vtkMatrixUtilities::ScalarTypeExtractor<MatrixT>::value_type Scalar;
 
@@ -1951,20 +2070,21 @@ inline void vtkQuaternionToMatrix3x3(const QuaternionT& quat, MatrixT& A)
 
   typedef vtkMatrixUtilities::Wrapper<3, 3, MatrixT> Wrapper;
 
-  Wrapper::template Get<0, 0>(A) = xx * f + s;
-  Wrapper::template Get<1, 0>(A) = (xy + wz) * f;
-  Wrapper::template Get<2, 0>(A) = (xz - wy) * f;
+  Wrapper::template Get<0, 0>(std::forward<MatrixT>(A)) = xx * f + s;
+  Wrapper::template Get<1, 0>(std::forward<MatrixT>(A)) = (xy + wz) * f;
+  Wrapper::template Get<2, 0>(std::forward<MatrixT>(A)) = (xz - wy) * f;
 
-  Wrapper::template Get<0, 1>(A) = (xy - wz) * f;
-  Wrapper::template Get<1, 1>(A) = yy * f + s;
-  Wrapper::template Get<2, 1>(A) = (yz + wx) * f;
+  Wrapper::template Get<0, 1>(std::forward<MatrixT>(A)) = (xy - wz) * f;
+  Wrapper::template Get<1, 1>(std::forward<MatrixT>(A)) = yy * f + s;
+  Wrapper::template Get<2, 1>(std::forward<MatrixT>(A)) = (yz + wx) * f;
 
-  Wrapper::template Get<0, 2>(A) = (xz + wy) * f;
-  Wrapper::template Get<1, 2>(A) = (yz - wx) * f;
-  Wrapper::template Get<2, 2>(A) = zz * f + s;
+  Wrapper::template Get<0, 2>(std::forward<MatrixT>(A)) = (xz + wy) * f;
+  Wrapper::template Get<1, 2>(std::forward<MatrixT>(A)) = (yz - wx) * f;
+  Wrapper::template Get<2, 2>(std::forward<MatrixT>(A)) = zz * f + s;
 }
 } // anonymous namespace
 
+VTK_ABI_NAMESPACE_BEGIN
 //------------------------------------------------------------------------------
 inline void vtkMath::QuaternionToMatrix3x3(const float quat[4], float A[3][3])
 {
@@ -1979,10 +2099,11 @@ inline void vtkMath::QuaternionToMatrix3x3(const double quat[4], double A[3][3])
 
 //-----------------------------------------------------------------------------
 template <class QuaternionT, class MatrixT, class EnableT>
-inline void vtkMath::QuaternionToMatrix3x3(const QuaternionT& q, MatrixT&& A)
+inline void vtkMath::QuaternionToMatrix3x3(QuaternionT&& q, MatrixT&& A)
 {
-  vtkQuaternionToMatrix3x3(q, A);
+  vtkQuaternionToMatrix3x3(std::forward<QuaternionT>(q), std::forward<MatrixT>(A));
 }
+VTK_ABI_NAMESPACE_END
 
 namespace
 {
@@ -1992,7 +2113,7 @@ namespace
 //  "Closed-form solution of absolute orientation using unit quaternions,"
 //  Journal of the Optical Society of America A, 4:629-642
 template <class MatrixT, class QuaternionT>
-inline void vtkMatrix3x3ToQuaternion(const MatrixT& A, QuaternionT& quat)
+inline void vtkMatrix3x3ToQuaternion(MatrixT&& A, QuaternionT&& quat)
 {
   typedef typename vtkMatrixUtilities::ScalarTypeExtractor<QuaternionT>::value_type Scalar;
 
@@ -2001,23 +2122,33 @@ inline void vtkMatrix3x3ToQuaternion(const MatrixT& A, QuaternionT& quat)
   typedef vtkMatrixUtilities::Wrapper<3, 3, MatrixT> Wrapper;
 
   // on-diagonal elements
-  N[0][0] = Wrapper::template Get<0, 0>(A) + Wrapper::template Get<1, 1>(A) +
-    Wrapper::template Get<2, 2>(A);
-  N[1][1] = Wrapper::template Get<0, 0>(A) - Wrapper::template Get<1, 1>(A) -
-    Wrapper::template Get<2, 2>(A);
-  N[2][2] = -Wrapper::template Get<0, 0>(A) + Wrapper::template Get<1, 1>(A) -
-    Wrapper::template Get<2, 2>(A);
-  N[3][3] = -Wrapper::template Get<0, 0>(A) - Wrapper::template Get<1, 1>(A) +
-    Wrapper::template Get<2, 2>(A);
+  N[0][0] = Wrapper::template Get<0, 0>(std::forward<MatrixT>(A)) +
+    Wrapper::template Get<1, 1>(std::forward<MatrixT>(A)) +
+    Wrapper::template Get<2, 2>(std::forward<MatrixT>(A));
+  N[1][1] = Wrapper::template Get<0, 0>(std::forward<MatrixT>(A)) -
+    Wrapper::template Get<1, 1>(std::forward<MatrixT>(A)) -
+    Wrapper::template Get<2, 2>(std::forward<MatrixT>(A));
+  N[2][2] = -Wrapper::template Get<0, 0>(std::forward<MatrixT>(A)) +
+    Wrapper::template Get<1, 1>(std::forward<MatrixT>(A)) -
+    Wrapper::template Get<2, 2>(std::forward<MatrixT>(A));
+  N[3][3] = -Wrapper::template Get<0, 0>(std::forward<MatrixT>(A)) -
+    Wrapper::template Get<1, 1>(std::forward<MatrixT>(A)) +
+    Wrapper::template Get<2, 2>(std::forward<MatrixT>(A));
 
   // off-diagonal elements
-  N[0][1] = N[1][0] = Wrapper::template Get<2, 1>(A) - Wrapper::template Get<1, 2>(A);
-  N[0][2] = N[2][0] = Wrapper::template Get<0, 2>(A) - Wrapper::template Get<2, 0>(A);
-  N[0][3] = N[3][0] = Wrapper::template Get<1, 0>(A) - Wrapper::template Get<0, 1>(A);
+  N[0][1] = N[1][0] = Wrapper::template Get<2, 1>(std::forward<MatrixT>(A)) -
+    Wrapper::template Get<1, 2>(std::forward<MatrixT>(A));
+  N[0][2] = N[2][0] = Wrapper::template Get<0, 2>(std::forward<MatrixT>(A)) -
+    Wrapper::template Get<2, 0>(std::forward<MatrixT>(A));
+  N[0][3] = N[3][0] = Wrapper::template Get<1, 0>(std::forward<MatrixT>(A)) -
+    Wrapper::template Get<0, 1>(std::forward<MatrixT>(A));
 
-  N[1][2] = N[2][1] = Wrapper::template Get<1, 0>(A) + Wrapper::template Get<0, 1>(A);
-  N[1][3] = N[3][1] = Wrapper::template Get<0, 2>(A) + Wrapper::template Get<2, 0>(A);
-  N[2][3] = N[3][2] = Wrapper::template Get<2, 1>(A) + Wrapper::template Get<1, 2>(A);
+  N[1][2] = N[2][1] = Wrapper::template Get<1, 0>(std::forward<MatrixT>(A)) +
+    Wrapper::template Get<0, 1>(std::forward<MatrixT>(A));
+  N[1][3] = N[3][1] = Wrapper::template Get<0, 2>(std::forward<MatrixT>(A)) +
+    Wrapper::template Get<2, 0>(std::forward<MatrixT>(A));
+  N[2][3] = N[3][2] = Wrapper::template Get<2, 1>(std::forward<MatrixT>(A)) +
+    Wrapper::template Get<1, 2>(std::forward<MatrixT>(A));
 
   Scalar eigenvectors[4][4], eigenvalues[4];
 
@@ -2039,6 +2170,7 @@ inline void vtkMatrix3x3ToQuaternion(const MatrixT& A, QuaternionT& quat)
 }
 } // anonymous namespace
 
+VTK_ABI_NAMESPACE_BEGIN
 //------------------------------------------------------------------------------
 inline void vtkMath::Matrix3x3ToQuaternion(const float A[3][3], float quat[4])
 {
@@ -2053,13 +2185,15 @@ inline void vtkMath::Matrix3x3ToQuaternion(const double A[3][3], double quat[4])
 
 //-----------------------------------------------------------------------------
 template <class MatrixT, class QuaternionT, class EnableT>
-inline void vtkMath::Matrix3x3ToQuaternion(const MatrixT& A, QuaternionT&& q)
+inline void vtkMath::Matrix3x3ToQuaternion(MatrixT&& A, QuaternionT&& q)
 {
-  vtkMatrix3x3ToQuaternion(A, q);
+  vtkMatrix3x3ToQuaternion(std::forward<MatrixT>(A), std::forward<QuaternionT>(q));
 }
+VTK_ABI_NAMESPACE_END
 
 namespace vtk_detail
 {
+VTK_ABI_NAMESPACE_BEGIN
 // Can't specialize templates inside a template class, so we move the impl here.
 template <typename OutT>
 void RoundDoubleToIntegralIfNecessary(double val, OutT* ret)
@@ -2091,8 +2225,10 @@ inline void RoundDoubleToIntegralIfNecessary(double val, float* retVal)
 
   *retVal = static_cast<float>(val);
 }
+VTK_ABI_NAMESPACE_END
 } // end namespace vtk_detail
 
+VTK_ABI_NAMESPACE_BEGIN
 //-----------------------------------------------------------------------------
 #if defined(VTK_HAS_ISINF) || defined(VTK_HAS_STD_ISINF)
 #define VTK_MATH_ISINF_IS_INLINE
@@ -2134,4 +2270,5 @@ inline bool vtkMath::IsFinite(double x)
 }
 #endif
 
+VTK_ABI_NAMESPACE_END
 #endif
